# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pandas

def read_data(filepath):
    """Imports the power spectrum data from the muon physics data file

    Arguments:
        filepath: The path to the .data file produced by the muon
                  acquisition software

    Returns:
        A 2D numpy array containing the data from the file. The first column 
        contains the acquisition intervals and the second column contains the
        timestamps.

    """

    df = pandas.read_csv(filepath, skiprows=0, header=None, sep=' ')
    return df.to_numpy()

def histc(x, bins):
    """Counts the number of times an values in :code:`x` fall between each bin in :code:`bins`.

    This uses the condition :code:`if bins[i] <= x[j] < bins[i+1]: result[i] += 1` 
    for each element in :code:`x`

    Arguments:
        x: The numpy array containing the data to count
        bins: A numpy array defining the bins

    Returns:
        A numpy array containing the number of elements between each bin.


    """
    indicies = np.digitize(x,bins)
    counts = np.zeros(bins.shape)
    for i in indicies:
        counts[i-1] += 1
    return counts


# Copyright 2019 School of Physics & Astronomy, Monash University),
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import pandas

def read_picoscope_csv(filepath):
    """Reads the csv file saved from the PicoScope software
    
    Arguments:
        filepath: A string containing the path to the csv file

    Returns:
        A tuple :code:`(time, CHA, CHB)` where each element is a 1D numpy 
        array containing an array of time 
        points (x-axis of the oscilloscope), and the y-values for channels A
        and B respectively. The units of these should be seconds, Volts and 
        Volts respectively unless otherwise stated via a warning message 
        printed to your terminal.
    """

    # read the data (skip the second line, and only use first 3 columns)
    df = pandas.read_csv(filepath, skiprows=[1,2], usecols=[0,1,2], na_values=['Infinity', '-Infinity'])

    time = df[df.columns[0]].values
    CHA = df[df.columns[1]].values
    CHB = df[df.columns[2]].values

    # now read the units
    df = pandas.read_csv(filepath, nrows=2, usecols=[0,1,2])
    
    time_unit = df[df.columns[0]].values[0]
    CHA_unit = df[df.columns[1]].values[0]
    CHB_unit = df[df.columns[2]].values[0]

    if time_unit == '(ms)':
        time = time * 1e-3
    elif time_unit == '(s)':
        pass
    elif time_unit == '(us)':
        time = time * 1e-6
    else:
        print('Warning: Unknown unit for time. Do not assume time is in seconds. It is in {}'.format(time_unit))

    if CHA_unit == '(mV)':
        CHA = CHA * 1e-3
    elif CHA_unit == '(V)':
        pass
    elif CHA_unit == ('uV'):
        CHA = CHA * 1e-6
    else:
        print('Warning: Unknown unit for CHA. Do not assume CHA is in Volts. It is in {}'.format(CHA_unit))

    if CHB_unit == '(mV)':
        CHB = CHB * 1e-3
    elif CHB_unit == '(V)':
        pass
    elif CHB_unit == ('uV'):
        CHB = CHB * 1e-6
    else:
        print('Warning: Unknown unit for CHB. Do not assume CHB is in Volts. It is in {}'.format(CHB_unit))

    return time, CHA, CHB


# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from .testing_helpers import compare_dictionary

def test_basic_linear_fit():
    """Basic test of linear fit without uncertainties"""

    ### Get results ###
    from monashspa.common.fitting import linear_fit, get_fit_parameters

    data = [
        [1,	2.97958155926148, 0.435952175570949],
        [2,	4.84845692653319, 0.467234424802873],
        [3,	6.44561871392239, 0.740203944022575],
        [4,	8.72544337668039, 0.509084442407005],
        [5,	10.7326357174404, 0.557439432218125],
        [6,	12.3202409005262, 0.0236735417569119],
    ]
    data = np.array(data)

    fit_result = linear_fit(data[:,0], data[:,1])
    results = get_fit_parameters(fit_result)

    ### Expected results ###
    expected_results = {
        'slope': 1.903875935,
        'u_slope': 0.045590672,
        'intercept': 1.011763758,
        'u_intercept': 0.177550159,
    }
    precision = 1e-7

    ### Check results match within precision ###
    success = compare_dictionary(results, expected_results, precision)

    return success


def test_linear_fit_with_uncertainties():
    """Basic test of linear fit with uncertainties"""

    ### Get results ###
    from monashspa.common.fitting import linear_fit, get_fit_parameters

    data = [
        [1,	2.97958155926148, 0.435952175570949],
        [2,	4.84845692653319, 0.467234424802873],
        [3,	6.44561871392239, 0.740203944022575],
        [4,	8.72544337668039, 0.509084442407005],
        [5,	10.7326357174404, 0.557439432218125],
        [6,	12.3202409005262, 0.0236735417569119],
    ]
    data = np.array(data)

    fit_result = linear_fit(data[:,0], data[:,1], u_y=data[:,2])
    results = get_fit_parameters(fit_result)

    ### Expected results (from WFIT) ###
    expected_results = {
        'slope': 1.866046353,
        'u_slope': 0.064841588,
        'intercept': 1.124423955,
        'u_intercept': 0.387570521,
    }
    precision = 2e-8

    ### Check results match within precision ###
    success = compare_dictionary(results, expected_results, precision)

    return success

def do_tests():
    tests = [test_basic_linear_fit, test_linear_fit_with_uncertainties]
    failed_tests = []
    print('Running common fitting tests...')

    for testfn in tests:
        print('    Running test "{test_name}":'.format(test_name=testfn.__doc__))
        result = testfn()
        print('        Result: {result}'.format(result='success' if result else 'failure'))

        if not result:
            failed_tests.append(testfn)

    if failed_tests:
        print('')
        print('    There were {num_failures:d} failed common fitting tests'.format(num_failures=len(failed_tests)))
        
    print('')

    return failed_tests

if __name__ == "__main__":
    do_tests()


# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

if __name__ == "__main__":
    # print python version
    import sys
    print("Python version:", sys.version)
    # print monashspa version
    import monashspa
    print("monashspa version:", monashspa.__version__)
    # print numpy version
    import numpy
    print("numpy version:", numpy.__version__)
    # print matplotlib version
    import matplotlib
    print("matplotlib version:", matplotlib.__version__)
    # print lmfit version
    import lmfit
    print("lmfit version:", lmfit.__version__)
    # print scipy version
    import scipy
    print("scipy version:", scipy.__version__)
    # print pandas version
    import pandas
    print("pandas version:", pandas.__version__)

    print()
    print("Running tests now...")
    
    failed_tests = []
    
    from . import fitting
    failed_tests.extend(fitting.do_tests())

    from .PHS2061 import fitting_tutorial
    failed_tests.extend(fitting_tutorial.do_tests())

    from .PHS3000 import fitting_tutorial as PHS3000_fitting_tutorial
    failed_tests.extend(PHS3000_fitting_tutorial.do_tests())

    from .PHS3000 import optical_tweezers
    failed_tests.extend(optical_tweezers.do_tests())

    print()
    print("Tests complete.")
    if failed_tests:
        print("Error: There were {} failed tests".format(len(failed_tests)))
    else:
        print("All tests completed successfully!")


# Copyright 2019 School of Physics & Astronomy, Monash University),
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

#: The calibration data for wavemeter #4929 as a 2D numpy array.
#: Here the first column corresponds to wavemeter reading in mm
#: and the second column corresponds to the frequency in GHz.
wavemeter_calibration = np.array([
    (24.215, 7.702),
    (22.675, 7.8),
    (21.255, 7.9),
    (19.97,  8.0),
    (18.81,  8.1),
    (17.75,  8.2),
    (16.79,  8.3),
    (15.91,  8.4),
    (15.06,  8.5),
    (14.36,  8.6),
    (13.64,  8.7),
    (12.98,  8.8),
    (12.35,  8.9),
    (11.77,  9.0),
    (11.21,  9.1),
    (10.69,  9.2),
    (10.20,  9.3),
    (9.73,   9.4),
    (9.29,   9.5),
    (8.87,   9.6),
    (8.46,   9.7),
    (8.08,   9.8),
    (7.71,   9.9),
    (7.36,   10.0),
    (7.03,   10.1),
    (6.71,   10.2),
    (6.39,   10.3),
    (6.09,   10.4),
    (5.81,   10.5),
    (5.52,   10.6),
    (5.26,   10.7),
    (5.00,   10.8),
    (4.75,   10.9),
    (4.50,   11.0),
    (4.28,   11.1),
    (4.06,   11.2),
    (3.84,   11.3),
    (3.62,   11.4),
    (3.42,   11.5),
    (3.22,   11.6),
    (3.03,   11.7),
    (2.85,   11.8),
    (2.66,   11.9),
    (2.49,   12.0),
])


# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import matplotlib.pyplot as plt
from matplotlib.legend import Legend as mplLegend

def savefig(fname, dpi=600, bbox_inches='tight', **kwargs):
    """ A wrapper for :py:func:`matplotlib.pyplot.savefig` with sensible defaults.
    
    By default, if a matplotlib legend is located outside of the plot axes,
    then :py:func:`matplotlib.pyplot.savefig` may cut off the legend when saving the
    figure. This function fixes this issue by setting :code:`bbox_inches='tight'`
    and setting :code:`bbox_extra_artists` to be a list of the current figure legends,
    unless the user chooses to define them otherwise. This function also sets 
    the default for :code:`dpi` to 600, which is large enough for most purposes.

    This function accepts any keyword argument from :py:func:`matplotlib.pyplot.savefig`
    and returns the result of the call to that function. 
    See :py:func:`matplotlib.pyplot.savefig` for usage.
    """
    
    if 'bbox_extra_artists' not in kwargs:
        ax = plt.gca()
        kwargs['bbox_extra_artists'] = [c for c in ax.get_children() if isinstance(c, mplLegend)]
    
    return plt.savefig(fname, dpi=dpi, bbox_inches=bbox_inches, **kwargs)


# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
try:
    from .__version__ import __version__
except ImportError:
    # Version file has not been autogenerated from build process:
    __version__ = 'Unknown development version'

def __colorise(enable, text):
    if enable:
        text = "\033[31;1m" +text+ "\033[0m"
    return text

def __check_for_new_monashspa_version():
    upgrade = False
    colorama_enabled = False
    try:
        import colorama
        colorama.init(strip=False)
        colorama_enabled = True
    except Exception:
        pass
    
    try:
        import requests
        # hack to work around the fact that sometimes the first attempt to 
        # reach PyPi times out (or takes much longer than it should), 
        # especially when the library is being imported for the first time 
        # in a while. Subsequent restarts of Python and importing don't
        # result in the bug occurring.
        # This might be a bug with the requests library, or it could be
        # something to do with DNS queries...not 100% sure.
        # Either way, initial testing indicates this works reasonably well.
        #
        # This code tries twice if it has to, raises the exception (to be
        # caught by the outer block) if it fails twice, and only checks once
        # if it succeeds first try.
        for i in range(2):
            try:
                r = requests.get('https://pypi.python.org/pypi/monashspa/json', timeout=0.1)
            except requests.exceptions.ReadTimeout:
                if i == 1:
                    raise
            else:
                if i == 0:
                    break
        
        # Get the latest version from PyPi
        online_version = r.json()['info']['version']
        if online_version != __version__:
            # check to see if version is newer
            o_major, o_minor, o_patch = online_version.split('.')
            try:
                l_major, l_minor, l_patch = __version__.split('.')
            except Exception:
                upgrade = True
            else:
                if int(l_major) < int(o_major):
                    upgrade = True
                elif int(l_major) == int(o_major) and int(l_minor) < int(o_minor):
                    upgrade = True
                elif int(l_major) == int(o_major) and int(l_minor) == int(o_minor) and int(l_patch) < int(o_patch):
                    upgrade = True
            
            if upgrade:
                print(__colorise(colorama_enabled, "[monashspa:info] A new version of monashspa is available"))
                print(__colorise(colorama_enabled, "[monashspa:info] You are running monashspa version: {}".format(__version__)))
                print(__colorise(colorama_enabled, "[monashspa:info] The latest monashspa version is: {}".format(online_version)))
                print(__colorise(colorama_enabled, "[monashspa:info] To upgrade, run: "))
                print(__colorise(colorama_enabled, "[monashspa:info]     pip install -U monashspa"))
                print(__colorise(colorama_enabled, "[monashspa:info] from a terminal or Anaconda Prompt (or see the install instructions at https://monashspa.readthedocs.io)"))
                print(__colorise(colorama_enabled, "[monashspa:info] You will need to restart your open Python/IPython/Spyder application after upgrading to use the new version."))
        
    except Exception:
        print(__colorise(colorama_enabled, "[monashspa:info] Unable to check if a newer version is available"))
        print(__colorise(colorama_enabled, "[monashspa:info] You are running monashspa version: {}".format(__version__)))
        print(__colorise(colorama_enabled, "[monashspa:info] You can find the latest version number at: https://monashspa.readthedocs.io"))
    
    try:
        colorama.deinit()
    except Exception:
        pass

    return upgrade

__check_for_new_monashspa_version()


# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import pandas
import numpy as np

def read_data(filepath):
    """Imports the 'xls' file (actually just a tsv) save by the Rigol Oscilloscope software

    Arguments:
        filepath: The path to the .xls or .txt file produced by the Rigol
                  oscilloscope software

    Returns:
        A tuple :code:`(t, V)` where :code:`t` is a 1D numpy array
        containing the time values associated with the oscilloscope trace
        and :code:`V` is a 1D numpy array of the same length with the associated
        voltage readings for each time point.

    """
    # TODO: consider replacing with our own csv reading wrapper
    df = pandas.read_csv(filepath, skiprows=list(range(7))+[8,], sep='\t')

    t = df[df.columns[1]].values
    V = df[df.columns[2]].values

    return t, V


# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

def compare_dictionary(results, expected_results, tolerance, indent=8):
    # check same number of results returned as expected
    success = len(expected_results) == len(results)
    
    # iterate over every result
    for k,v in results.items():
        # check key names match
        if k not in expected_results:
            success = False
            print(' '*indent + 'Key "{key}" missing from expected results'.format(key=k))
        # Check result is within tolerance to expected value
        if np.abs(expected_results[k] - v) > tolerance:
            success = False
            print(' '*indent + 'Values for "{key}" are not within tolerance.'.format(key=k))
            print(' '*(indent+4) + 'Expected value: {}'.format(expected_results[k]))
            print(' '*(indent+4) + 'Actual value: {}'.format(v))
            print(' '*(indent+4) + 'Tolerance: {}'.format(tolerance))
            print(' '*(indent+4) + 'Difference: {}'.format(np.abs(expected_results[k] - v)))

    # this check technically isn't needed because we compare length and check if
    # all of one is in the other, however this allows us to provide more useful
    # feedback on failed tests
    for k,v in expected_results.items():
        if k not in results:
            success = False
            print(' '*indent + 'Key "{key}" missing from results'.format(key=k))

    return success


# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from ..testing_helpers import compare_dictionary

def nonlinear_fit():
    """Basic test of PHS3000 nonlinear fit"""

    ### Get results ###
    import monashspa.PHS3000 as spa

    # load the data
    data = spa.tutorials.fitting.part_b_data
    # slice the data into columns
    t = data[:,0]
    A = data[:,1]
    u_A = data[:,2]

    model = spa.make_lmfit_model("A_0*exp(-l*x)")
    params = model.make_params(A_0=30, l=.005)
    params.add('halflife', expr="log(2)/l")
    fit_results = spa.model_fit(model, params, x=t, y=A, u_y=u_A)
    results = spa.get_fit_parameters(fit_results)

    ### Expected results ###
    expected_results = {
        'A_0': 20.573035990441486,
        'u_A_0': 0.6181473325960677,
        'l': 0.005004779941925933,
        'u_l': 0.00015840653659960648,
        'halflife': 138.49703455557113,
        'u_halflife': 4.38357646646529,
    }
    precision = 2e-7
    
    ### Check results match within precision ###
    success = compare_dictionary(results, expected_results, precision)

    return success

def nonlinear_fit_with_independent_as_t():
    """Test of PHS3000 nonlinear fit using an independent variable that was not 'x'"""

    ### Get results ###
    import monashspa.PHS3000 as spa

    # load the data
    data = spa.tutorials.fitting.part_b_data
    # slice the data into columns
    t = data[:,0]
    A = data[:,1]
    u_A = data[:,2]

    model = spa.make_lmfit_model("A_0*exp(-l*t)", independent_vars=["t"])
    params = model.make_params(A_0=30, l=.005)
    params.add('halflife', expr="log(2)/l")
    fit_results = spa.model_fit(model, params, x=t, y=A, u_y=u_A)
    results = spa.get_fit_parameters(fit_results)

    ### Expected results ###
    expected_results = {
        'A_0': 20.573035990441486,
        'u_A_0': 0.6181473325960677,
        'l': 0.005004779941925933,
        'u_l': 0.00015840653659960648,
        'halflife': 138.49703455557113,
        'u_halflife': 4.38357646646529,
    }
    precision = 2e-7

    ### Check results match within precision ###
    success = compare_dictionary(results, expected_results, precision)

    return success

def linear_fit_model():
    """Test of PHS3000 linear fit using custom model"""

    ### Get results ###
    import monashspa.PHS3000 as spa

    # load the data
    data = spa.tutorials.fitting.part_b_data
    # slice the data into columns
    t = data[:,0]
    A = data[:,1]
    u_A = data[:,2]

    model = spa.make_lmfit_model("m*x+c")
    params = model.make_params(m=0.04, c=5)
    params.add('halflife', expr="-log(2)/m")
    params.add('A_0', expr="exp(c)")
    fit_results = spa.model_fit(model, params, x=t, y=np.log(A), u_y=u_A/A)
    results = spa.get_fit_parameters(fit_results)

    ### Expected results ###
    expected_results = {
        'c': 3.02843895796932,
        'u_c': 0.030620495054301592,
        'm': -0.004897938859868103,
        'u_m': 0.0001633020007207088,
        'halflife': 141.51813658584769,
        'u_halflife': 4.718351025589936,
        'A_0': 20.664948544330286,
        'u_A_0': 0.6327709546990624,
    }
    precision = 1e-5

    ### Check results match within precision ###
    success = compare_dictionary(results, expected_results, precision)

    return success

def linear_fit():
    """Basic test of PHS3000 linear fit"""

    ### Get results ###
    import monashspa.PHS3000 as spa

    # load the data
    data = spa.tutorials.fitting.part_b_data
    # slice the data into columns
    t = data[:,0]
    A = data[:,1]
    u_A = data[:,2]

    fit_results = spa.linear_fit(t, np.log(A), u_y=u_A/A)
    results = spa.get_fit_parameters(fit_results)

    ### Expected results ###
    expected_results = {
        'intercept': 3.02843895796932,
        'u_intercept': 0.030620495054301592,
        'slope': -0.004897938859868103,
        'u_slope': 0.0001633020007207088,
    }
    precision = 1e-7

    ### Check results match within precision ###
    success = compare_dictionary(results, expected_results, precision)

    return success

def linear_fit_dual_custom_model():
    """Test of PHS3000 linear fit using the sum of two custom models"""

    ### Get results ###
    import monashspa.PHS3000 as spa

    # load the data
    data = spa.tutorials.fitting.part_b_data
    # slice the data into columns
    t = data[:,0]
    A = data[:,1]
    u_A = data[:,2]

    model1 = spa.make_lmfit_model("m*x")
    model2 = spa.make_lmfit_model("c")
    model = model1 + model2
    params = model.make_params(m=0.04, c=5)
    params.add('halflife', expr="-log(2)/m")
    params.add('A_0', expr="exp(c)")
    fit_results = spa.model_fit(model, params, x=t, y=np.log(A), u_y=u_A/A)
    results = spa.get_fit_parameters(fit_results)

    ### Expected results ###
    expected_results = {
        'c': 3.02843895796932,
        'u_c': 0.030620495054301592,
        'm': -0.004897938859868103,
        'u_m': 0.0001633020007207088,
        'halflife': 141.51813658584769,
        'u_halflife': 4.718351025589936,
        'A_0': 20.664948544330286,
        'u_A_0': 0.6327709546990624,
    }
    precision = 1e-5

    ### Check results match within precision ###
    success = compare_dictionary(results, expected_results, precision)

    return success

def linear_fit_dual_custom_named_model():
    """Test of PHS3000 linear fit using the sum of two custom models with prefixes"""

    ### Get results ###
    import monashspa.PHS3000 as spa

    # load the data
    data = spa.tutorials.fitting.part_b_data
    # slice the data into columns
    t = data[:,0]
    A = data[:,1]
    u_A = data[:,2]

    model1 = spa.make_lmfit_model("m*x", prefix="m1")
    model2 = spa.make_lmfit_model("c", prefix="m2")
    model = model1 + model2
    params = model.make_params(m1m=0.04, m2c=5)
    params.add('halflife', expr="-log(2)/m1m")
    params.add('A_0', expr="exp(m2c)")
    fit_results = spa.model_fit(model, params, x=t, y=np.log(A), u_y=u_A/A)
    results = spa.get_fit_parameters(fit_results)

    ### Expected results ###
    expected_results = {
        'm2c': 3.02843895796932,
        'u_m2c': 0.030620495054301592,
        'm1m': -0.004897938859868103,
        'u_m1m': 0.0001633020007207088,
        'halflife': 141.51813658584769,
        'u_halflife': 4.718351025589936,
        'A_0': 20.664948544330286,
        'u_A_0': 0.6327709546990624,
    }
    precision = 1e-5

    ### Check results match within precision ###
    success = compare_dictionary(results, expected_results, precision)

    return success

def non_linear_fit_part_c():
    """Test of PHS3000 non-linear fit to multi-model data"""

    ### Get results ###
    import monashspa.PHS3000 as spa

    # load the data
    data = spa.tutorials.fitting.part_c1_data
    # slice the data into columns
    x = data[:,0]
    y = data[:,1]

    # Define the models
    model1 = spa.make_lmfit_model("A1*exp(-(x-B1)**2/D1**2)", name="Gaussian1")
    model2 = spa.make_lmfit_model("A2*exp(-(x-B2)**2/D2**2)", name="Gaussian2")
    model3 = spa.make_lmfit_model("c+x*0", name="Offset")
    model = model1 + model2 + model3
    params = model.make_params(A1=np.max(y), B1=100, D1=6, A2=np.max(y), B2=160, c=3)
    params.add('D2', expr='D1')
    params.add('FWHM', expr='2*sqrt(log(2)*D1)')
    fit_results = spa.model_fit(model, params, x=x, y=y)
    results = spa.get_fit_parameters(fit_results)

    ### Expected results ###
    expected_results = {
        'A1': 16.58279101985785, 
        'u_A1': 0.351003396364092, 
        'B1': 95.82492417888295, 
        'u_B1': 0.12363799249150613, 
        'D1': 6.705149651134903, 
        'u_D1': 0.1412232821933916, 
        'A2': 13.690661831671, 
        'u_A2': 0.33945028545509237, 
        'B2': 164.19308257427863, 
        'u_B2': 0.1497567729486557, 
        'D2': 6.705149651134903, 
        'u_D2': 0.14122328218157232, 
        'c': 2.636117481366362, 
        'u_c': 0.06764023986252798, 
        'FWHM': 4.311684392863958, 
        'u_FWHM': 0.045406161696634036
    }
    precision = 1e-5

    ### Check results match within precision ###
    success = compare_dictionary(results, expected_results, precision)

    return success

def do_tests():
    tests = [nonlinear_fit, nonlinear_fit_with_independent_as_t, linear_fit, linear_fit_model, linear_fit_dual_custom_model, linear_fit_dual_custom_named_model, non_linear_fit_part_c]
    failed_tests = []

    print('Running PHS3000 fitting tutorial tests...')

    for testfn in tests:
        print('    Running test "{test_name}":'.format(test_name=testfn.__doc__))
        result = testfn()
        print('        Result: {result}'.format(result='success' if result else 'failure'))

        if not result:
            failed_tests.append(testfn)

    if failed_tests:
        print('')
        print('    There were {num_failures:d} failed PHS3000 fitting tests'.format(num_failures=len(failed_tests)))
    print('')

    return failed_tests

if __name__ == "__main__":
    do_tests()


# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

def read_mca_file(filepath):
    """Imports data saved from the ADMCA software

    Arguments:
        filepath: The path to the .mca file produced by the ADMCA
                  acquisition software

    Returns:
        A tuple :code:`(header, data)` where :code:`header` is a dictionary
        containing the key, value pairs from header rows of the .mca file and
        :code:`data` is a 1D numpy array of the counts for each MCA channel.

    """

    header = {}
    data = []
    reading_header = False
    reading_data = False
    with open(filepath, 'r') as f:
        for line in f:
            line = line.strip('\r\n')
            if line == '<<PMCA SPECTRUM>>':
                reading_header = True
                continue
            elif line == '<<DATA>>':
                reading_header = False
                reading_data = True
                continue
            elif line == '<<END>>':
                break
            elif reading_header:
                try:
                    key, value = line.split(' - ')
                    header[key] = value
                except Exception:
                    pass
            elif reading_data:
                data.append(int(line))
    return header, np.array(data)



# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import os
from setuptools import setup, find_packages

# Define the current version of the library
# Update this before building and publishing a new version
# see https://semver.org/ for guidelines on how to modify the version string
VERSION = '1.1.0'

# get directory of setup.py and the rest of the code for the library
code_dir = os.path.abspath(os.path.dirname(__file__))

# Auto generate a __version__ file for the package to import
with open(os.path.join(code_dir, 'monashspa', '__version__.py'), 'w') as f:
    f.write("__version__ = '%s'\n" % VERSION)

# Work around the fact that the readme.md file doesn't exist for users installing
# from the tar.gz format. However, in this case, they won't be uploading to PyPi
# so they don't need it!
try:
    # Read in the readme file as the long description
    with open(os.path.join(code_dir, 'readme.md')) as f:
        long_description = f.read()
except Exception:
    long_description = ""

setup(
    name='monashspa',
    version=VERSION,
    description='Library of useful data analysis tools for Monash University Physics & Astronomy students',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://monashspa.readthedocs.io',
    project_urls={
        "Documentation": "https://monashspa.readthedocs.io",
        "Source Code": "https://bitbucket.org/monashuniversityphysics/monashspa",
    },
    license='GPLv3',
    author='School of Physics & Astronomy, Monash University',
    classifiers=['Development Status :: 5 - Production/Stable',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3.4',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7',
                 'Environment :: Console',
                 'Intended Audience :: Education',
                 'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
                 'Natural Language :: English',
                 'Operating System :: OS Independent',
                ],
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'numpy>=1.15',
        'lmfit>=0.9.12,<1.0',
        'requests>=2.21.0,<3',
        'colorama>=0.4.1,<1',
        'pandas>=0.24,<1',
        'six',
        'piradon',
    ],
    data_files=[
        ('monashspa/PHS2061', ['monashspa/PHS2061/PHS20x1UncertaintiesData.csv']),
    ]
)


# Copyright 2019 School of Physics & Astronomy, Monash University
#
# This file is part of monashspa.
#
# monashspa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# monashspa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with monashspa.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

#: Resistance data for temperatures below 77K as a 2D numpy array.
#: Columns are: temperature (K), u_temperature (K), resistance (Ohms),
#: u_resistance(Ohms) respectively
liquid_helium = np.array([
    (76.9, 0.1, 37,   1),
    (67.0, 0.1, 32,   1),
    (61.5, 0.1, 30,   1),
    (57.0, 0.1, 29,   1),
    (51.9, 0.1, 27,   1),
    (47.0, 0.1, 26,   1),
    (42.0, 0.1, 26,   1),
    (38.0, 0.1, 26,   1),
    (31.9, 0.1, 28,   1),
    (26.8, 0.1, 31,   1),
    (21.9, 0.1, 39,   1),
    (20.0, 0.1, 46,   1),
    (18.1, 0.1, 53,   1),
    (16.0, 0.1, 65,   1),
    (13.8, 0.1, 106,  1),
    (12.0, 0.1, 162,  1),
    (10.0, 0.1, 526,  1),
    (8.0,  0.1, 3.23, 0.01e3),
    (6.0,  0.1, 292,  1e3),
    (4.0,  0.1, 81.7, 0.1e6),
])
